#ifndef	   MY_CONDITIONS_H_INCLUDED
#define	   MY_CONDITIONS_H_INCLUDED

#include "d2_straight_line.h"
#include "field_predicate.h"


#include "autogen_act_single_hard_kick_to.h"

////////////////////////////////////////////////////////////////////////
//
//	Pass Ratio Check
//
////////////////////////////////////////////////////////////////////////
class Pass_Ratio_Check : public Field_Predicate
{
public:
	enum Pass_Type {
		Type_None,
		Type_Center,
		Type_Side,
		Type_Forward,
		Type_Backward,
		Type_Rside ,
		Type_Lside ,
		Type_Ownside ,
		Type_Near ,
		Type_Far };

public:
	virtual ~Pass_Ratio_Check(){}
	virtual bool	operator() ( const Field_Recog_Interface & f ) const
	{
		if ( get_pass_unum(f) != -1 )
			return true ;
		else
			return false ;
	}

protected:
	Pass_Type	ptype ;

	mutable SServer_Time set_unum_time ;
	mutable int pass_unum ;

public:
	Pass_Ratio_Check ( Pass_Type set_type = Type_None ) ;
	void set_pass_type ( Pass_Type set_type ) ;
	void set_pass_unum ( const Field_Recog_Interface & f,const int i ) const ;
	int get_pass_unum ( const Field_Recog_Interface & f ) const ;
	int get_pass_teammate_unum ( const Field_Recog_Interface & f ) const ;
	double  pass_check_angle3( const Field_Recog_Interface &  f ,
				   const D2_Vector &  from , const D2_Vector &  to ) const ;
	bool pass_check_angle2 ( const Field_Recog_Interface & f,const D2_Vector & target , double ratio ) const ;
	bool pass_check_angle ( const Field_Recog_Interface & f,const D2_Vector & target , double ratio ) const ;
	bool pass_check_danger ( const Field_Recog_Interface & f,const D2_Vector & target ) const ;
	bool pass_check_type ( const Field_Recog_Interface & f,const D2_Vector & target , Pass_Type pass_type ) const ;
	bool pass_finished ( const Field_Recog_Interface & f ) const ;
} ;

class PassSpace : public Field_Predicate
{
public:
	enum PassCheckType { Angle_Check , Coordinate_Check } ;

protected:
	mutable bool		pass_enable ;
	mutable SServer_Time	check_time ;
	mutable D2_Vector	target ;
	mutable D2_Vector	tmp_target  ;

	Command_Queue	q;
	PassCheckType	ptype ;
public:

	virtual ~PassSpace() {}
	virtual bool operator() ( const Field_Recog_Interface & f ) const
	{
		if ( check_pass_enable ( f ) == true )
			return true ;
		else
			return false ;
	}

	PassSpace(PassCheckType p = Angle_Check)
		:ptype(p)
	{
		check_time = -1 ;
		pass_enable = false ;
		
	}

	void	set_pass_check_type ( PassCheckType p ) ;
	bool	check_pass_enable ( const Field_Recog_Interface & f ) const ;
	D2_Vector	pass_point ( const Field_Recog_Interface & f ) const ;
	PassCheckType	pass_type ( const Field_Recog_Interface & f ) const ;
	bool pass_receive_enable_teammate ( const Field_Recog_Interface & f , const Angle & pass_abs_angle ) const ;
	bool pass_receive_enable_teammate ( const Field_Recog_Interface & f , const D2_Vector & pass_point ) const ;
	void	calc_pass_point_angle ( const Field_Recog_Interface & f ) const ;
	void	calc_pass_point_vector ( const Field_Recog_Interface & f ) const ;
	void	calc_pass_point ( const Field_Recog_Interface & f ) const ;
} ;

class ThroughPass : public Field_Predicate
{
protected:

	mutable bool		pass_enable ;
	mutable SServer_Time	check_time ;
	mutable D2_Vector	target ;

	Command_Queue	q;

public:
	virtual ~ThroughPass() {}
	virtual bool operator() ( const Field_Recog_Interface & f ) const
	{
		if ( check_pass_enable ( f ) == true )
			return true ;
		else
			return false ;
	}

	bool	check_pass_enable ( const Field_Recog_Interface & f ) const
	{
		if ( f.current_time() != check_time )
		{
			calc_pass_point ( f ) ;
		}
		return pass_enable ;
	}

	D2_Vector	pass_point ( const Field_Recog_Interface & f ) const
	{
		if ( f.current_time() != check_time )
		{
			calc_pass_point ( f ) ;
		}
		return target ;
	}

	bool	pass_receive_enable ( const Field_Recog_Interface & f , const D2_Vector & pass_point , const Player_Reference & receiver ) const
	{
		//ѥѥݥȤ˸ƽФȤơreceiverܡ뤫ɤ
		D2_Vector predict_ball_velocity ;
		predict_ball_velocity = Act_Single_Hard_Kick_To::result(f,pass_point) ;

		D2_Vector ball_init_point = f.ball().coordinate() ;

		const int MAX_SEACH_STEP = 50 ;

		//Ȥꤢη׻̵
		int turn_step = 0 ;

		for ( int t = 0 ; t < MAX_SEACH_STEP ; t ++ )
		{
			//ԥΥܡΰ
			D2_Vector ball_current_point = f.sserver_param().ball_basic_point_n_step_later( ball_init_point,
													predict_ball_velocity,
													t ) ;
			//ܡ뤬fieldγ˹ԤäƤޤäΤʤ
			if ( fabs(ball_current_point.x()) > f.sserver_param().field_half_length()
			     || fabs(ball_current_point.y()) > f.sserver_param().field_half_width() )
				return false ;

			//ԥǰưǽʵΥ
			//D2_Vector player_movable_point
			//	= f.sserver_param().player_basic_point_n_step_later(D2_Vector(0,0),D2_Vector(0,0),(t-turn_step)) ;
			D2_Vector player_movable_point ;
			player_movable_point = D2_Vector ( f.sserver_param().PLAYER_SPEED_MAX()*(t-turn_step),0) ;

			double movable_distance = player_movable_point.r() ;
			double point_to_point_distance = (ball_current_point - receiver.coordinate() ).r() ;

			if ( point_to_point_distance < movable_distance )
			{
				if ( dbg )
				{
					dbg << "++++++++++++++++++++" << endl ;
					dbg << "Pass Receive Enable!" << endl ;
					dbg << "++++++++++++++++++++" << endl ;
					dbg << "Pass Receive Point ("<< t << ") [" << ball_current_point << "]" << endl ;
					dbg << "Turn = " << turn_step << endl ; 
					dbg << "movable distance ("<< t << ") [" << movable_distance << "]" << endl ;
					dbg << "ball <--> receiver distance ("<< t << ") [" << point_to_point_distance << "]" << endl ;
				}

				return true ;
			}
		}
		return false ;
	}

	void	calc_pass_point ( const Field_Recog_Interface & f ) const
	{
		//ѥΥåȤܤƥݥȤ򸫤Ĥ
		//offence_line˶ᤤȡ
		//offence_line΢˥ڡ뤳ȡ
		//ѥ뤳ȡ

		D2_Vector		best_point ;
		double			best_angle = -1000 ;

		for ( int i = 1 ; i <= MAX_PLAYER ; ++ i )
		{
			//ѡȼʬоݳ
			if ( f.teammate(i).self_player() ) continue ;
			if ( f.teammate(i).goalie() ) continue ;

			//10˸ƤΤʤ顢åоݤȤ
			if ( ! f.teammate(i).coordinate().accuracy_check(-10) ) continue ;

			if ( fabs(f.teammate(i).coordinate().x() - f.our_team().offence_line()) > 8.0
			     ||
			     f.our_team().offence_line() > f.teammate(i).coordinate().x()
			     )
				continue;

			D2_Vector check_base_point = f.teammate(i).coordinate()
				+ D2_Vector ( (f.teammate(i).coordinate() - f.ball().coordinate()).r()*0.1+6.0 ,0) ;

			if ( dbg )
			{
				dbg << "Throught pass player number = " << i << endl ;
				dbg << "Throught pass player coordinate = " << f.teammate(i).coordinate() << endl ;
				dbg << "Pass target base point ? " << check_base_point << endl ;
			}

			Pass_Ratio_Check	pcheck ;
			pcheck.set_pass_type(Pass_Ratio_Check::Type_Forward) ;
			//int target_unum = pcheck.get_pass_unum(f) ;

			for ( double x = -3.0 ; x <= 15.0 ; x+=2.0 )
			{
				for ( double y = -5.0 ; y <= 5.0 ; y+=2.0 )
				{
					D2_Vector pass_check_point(check_base_point.x()+x,check_base_point.y()+y) ;

					if ( pcheck.pass_check_type ( f,pass_check_point , Pass_Ratio_Check::Type_Forward ) )
					{
						double min_pass_angle = pcheck.pass_check_angle3(f,f.ball().coordinate(),pass_check_point) ;

						if ( pass_receive_enable ( f , pass_check_point , f.teammate(i) ) == false )
						{
							if ( dbg )
							{
								dbg << "This point is disable [" << pass_check_point << "]" << endl ;
							}
							continue ;
						}

						if ( best_angle <= min_pass_angle )
						{
							if ( dbg )
							{
								dbg << "Update Best angle " << pass_check_point
								    << "angle(" << min_pass_angle << ")" << endl ;
							}
							best_angle = min_pass_angle ;
							best_point = pass_check_point ;
						}
					}
				}
			}
		}

		if ( best_angle < 10.0 )
		{
			pass_enable = false ;
			check_time = f.current_time() ;
		}
		else
		{
			pass_enable = true ;
			target = best_point ;
			check_time = f.current_time() ;
			if ( dbg )
			{
				dbg << "-------------------------------------------------" << endl ;
				dbg << "Time = " << check_time.main_step() << endl ;
				dbg << "Best Through Pass Point = " << best_point << endl ;
				dbg << "-------------------------------------------------" << endl ;
			}
		}
	}
} ;

class ForwardSpace : public Field_Predicate
{
public:
	enum Space_Type {
		Type_Dribble,
		Type_None
	} ;


public:
	virtual ~ForwardSpace() {}
	virtual bool operator() ( const Field_Recog_Interface & f ) const
	{
		if ( check_space ( f ) == true )
			return true ;
		else
			return false ;
	}

protected:
	Space_Type space_type ;
	mutable SServer_Time check_time ;
	mutable	bool check_result ;

public:
	ForwardSpace ( Space_Type set_type = Type_None )
		:space_type(set_type)
	{
		check_time = -1 ;
		check_result = false ;
	}

	void set_space_type ( Space_Type set_type )
	{
		if ( space_type != set_type )
		{
			space_type = set_type ;
			check_result = -1 ;
		}
	}

	// (tmp)
	void set_result ( const Field_Recog_Interface & f,const bool result ) const
	{
		check_result = result ;
		check_time = f.current_time() ;
	}

	bool check_space ( const Field_Recog_Interface & f ) const
	{
		if ( this -> check_time == f.current_time() )
		{
			return check_result ;
		}

		//check forward space
		check_result = check_forward_space ( f ) ;
		check_time = f.current_time() ;
		return check_result ;
	}

	bool check_forward_space ( const Field_Recog_Interface & f ) const
	{
		const double max_check_dist = 30.0 ;
		const Angle max_check_angle = Degree(90.0) ;

		double check_dist = 0 ;
		Angle  check_angle = Angle(Angle::Degree,0.0) ;

		D2_Vector my_position = f.self().coordinate() ;

		if ( space_type == Type_Dribble )
		{
			check_dist = 15.0 ;
			check_angle = Degree(50.0) ;
		}
		else if ( space_type == Type_None )
		{
			check_dist = max_check_dist ;
			check_angle = max_check_angle ;
		}
		else
		{
			check_dist = max_check_dist ;
			check_angle = max_check_angle ;
		}

		//
		//   forward
		//	|
		//	|
		//	|	target(x,y)
		//	|
		//	|
		//     self

		double  kickable_dist = f.sserver_param().kickable_distance() -0.05 ;

		D2_Straight_Line line( f.self().coordinate() , f.self().coordinate() + D2_Vector(check_dist,0) ) ;

		D2_Vector pt ;
		D2_Vector proj_pt ;

		for ( int i = 1 ; i <= MAX_PLAYER ; ++ i )
		{
			if ( f.opponent(i).coordinate().accuracy_check(-5) )
			{
				pt = f.opponent(i).coordinate() ;

				try
				{
					proj_pt = line.project_point(pt) ;
				}
				catch (D2_Region::No_Region_Error)
				{
					continue ;
				}

				D2_Vector x_l ;
				D2_Vector y_l ;

				x_l = proj_pt - f.self().coordinate() ;
				y_l = (proj_pt - pt) *
					( ( (proj_pt-pt).r() - kickable_dist ) / ( proj_pt-pt).r() ) ;

				if ( x_l.r() == 0 || x_l.r() > check_dist )
					continue ;

				if ( check_angle < Degree(fabs( atan ( y_l.r()/x_l.r() ) )) )
				{
					continue ;
				}
				else
				{
					return false ;
				}
			}
		}

		for ( size_t i = 0 ; i < f.n_unknown_opponent() ; ++ i )
		{
			if ( f.unknown_opponent(i).coordinate().accuracy_check(-5) )
			{
				pt = f.unknown_opponent(i).coordinate() ;

				try
				{
					proj_pt = line.project_point(pt) ;
				}
				catch (D2_Region::No_Region_Error)
				{
					continue ;
				}

				D2_Vector x_l ;
				D2_Vector y_l ;

				x_l = proj_pt - f.self().coordinate() ;
				y_l = (proj_pt - pt) *
					( ( (proj_pt-pt).r() - kickable_dist ) / ( proj_pt-pt).r() ) ;

				if ( x_l.r() == 0 || x_l.r() > check_dist )
					continue ;

				if ( check_angle < Degree(fabs( atan ( y_l.r()/x_l.r() ) )) )
				{
					continue ;
				}
				else
				{
					return false ;
				}
			}
		}

		for ( size_t i = 0 ; i < f.n_unknown_player() ; ++ i )
		{
			if ( f.unknown_player(i).coordinate().accuracy_check(-5) )
			{
				pt = f.unknown_player(i).coordinate() ;

				try
				{
					proj_pt = line.project_point(pt) ;
				}
				catch (D2_Region::No_Region_Error)
				{
					continue ;
				}

				D2_Vector x_l ;
				D2_Vector y_l ;

				x_l = proj_pt - f.self().coordinate() ;
				y_l = (proj_pt - pt) *
					( ( (proj_pt-pt).r() - kickable_dist ) / ( proj_pt-pt).r() ) ;

				if ( x_l.r() == 0 || x_l.r() > check_dist )
					continue ;

				if ( check_angle < Degree(fabs( atan ( y_l.r()/x_l.r() ) )) )
				{
					continue ;
				}
				else
				{
					return false ;
				}
			}
		}

		return true ;
	}
} ;

////////////////////////////////////////////////////////////////////////
//
//	Ball Kickable
//
////////////////////////////////////////////////////////////////////////
class BallKickable : public Field_Predicate
{
public:
	virtual ~BallKickable(){}
	virtual bool	operator() ( const Field_Recog_Interface & f ) const
	{
		if ( (f.ball().coordinate() - f.self().coordinate()).r()
		     > f.sserver_param().kickable_distance() - 0.05 )
			return false ;
		else
			return true ;
	}
} ;


#include <cfloat>

class NotDanger : public Field_Predicate
{
public:
	virtual ~NotDanger(){}

	virtual bool	operator() ( const Field_Recog_Interface & f ) const
	{
		Player_Reference	closest_player( f );
		double			min_dist = DBL_MAX;

		for( int  i = 1  ;  i <= MAX_PLAYER  ;  i ++ )
		{
			if ( ! f.opponent(i).coordinate().accuracy_check( -10 ) )
			{
				continue;
			}

			if ( dbg )
			{
				//dbg << "Opp(" << i << ") position is valid" << endl ;
			}
			double	dist = (f.opponent(i).coordinate()
					- f.self().coordinate()).r() ;
			if ( dbg )
			{
				//dbg << "Opp(" << i << ") dist             (" << dist << ")" << endl ;
			}

			if ( dist < min_dist )
			{
				min_dist = dist;
				if ( dbg )
				{
					//dbg << "Current Min dist = " << min_dist << endl ;
				}
				closest_player = f.opponent(i) ;
				if ( dbg )
				{
					if ( dbg )
					{
						dbg << "Closest Distance is ("
						    << (f.self().coordinate() - f.opponent(i).coordinate()).r()
						    << ")" << endl ;
					}
				}
			}
		}

		for( size_t  i = 0  ;  i < f.n_unknown_opponent()  ;  i ++ )
		{
			double	dist = (f.unknown_opponent(i).coordinate()
					- f.self().coordinate()).r();

			if ( dbg )
			{
				dbg << "Unknown(" << i << ") position is valid" << endl ;
			}
			if ( dist < min_dist )
			{
				min_dist = dist;
				if ( dbg )
				{
					dbg << "Current Min dist = " << min_dist << endl ;
				}
				closest_player = f.unknown_opponent(i);
			}
		}

		for( size_t  i = 0  ;  i < f.n_unknown_player()  ;  i ++ )
		{
			double	dist = (f.unknown_player(i).coordinate()
					- f.self().coordinate()).r();

			if ( dbg )
			{
				dbg << "Unknown(" << i << ") position is valid" << endl ;
			}
			if ( dist < min_dist )
			{
				min_dist = dist;
				if ( dbg )
				{
					dbg << "Current Min dist = " << min_dist << endl ;
				}
				closest_player = f.unknown_player(i);
			}
		}

		if ( closest_player.valid() )
		{
			D2_Vector	closest_player_coordinate
				= closest_player.coordinate()
				+ closest_player.velocity();

#if 1
			D2_Vector	rel_player
				= (closest_player_coordinate
				   - f.self().coordinate())
				.rotate( - f.self().body_angle() ) ;
			if ( dbg )
			{
				dbg << "Closest Distance is (" << rel_player.r() << ")" << endl ;
				dbg << "Closest Distance is (" << (f.self().coordinate() - closest_player_coordinate).r() << ")" << endl ;
			}
#endif
			if ( rel_player.r() < 6.5 )
			//if ( (f.self().coordinate() - closest_player_coordinate).r() < 6.0  )
			{
				if ( dbg )
				{
					dbg << "Time = " << f.current_time().main_step() << endl ;
					dbg << "Danger Danger Danger!!" << endl ;
				}
				return false ;
			}
		}

		if ( dbg )
		{
			dbg << "Time = " << f.current_time().main_step() << endl ;
			dbg << "Not danger Not danger Not danger!!" << endl ;
		}
		return true ;
	}

};
#endif
