#ifndef	   PASS_COURCE_TABLE_H_INCLUDED
#define	   PASS_COURCE_TABLE_H_INCLUDED

// Author:		H. Shimora
// Created:		Feb 16 2001
// Version:		0.00

///-----------------------------------------------
/// Change Log:
///-----------------------------------------------
// version 0.00  Feb 16 2001    base version.
//
//

#include  <vector>
#include  "field_recog_interface.h"
#include  "soccer_action.h"
#include  "ref_count_ptr.h"

class  Pass_Cource_Table
{
protected:
	static	const	int	RECURSIVE_LEVEL = 5;

	typedef  int  State;
	static	const	int	PUSH_STATE = MAX_PLAYER + 1;
	static	const	int	GOAL_STATE = MAX_PLAYER + 2;

	static	const	double	PASS_EVAL_THRESHOLD;

	struct  Connection_Table
	{
	public:
		std::vector<double>	player;
		bool			goal;
		bool			push;

		double			free_radius;

	public:
		double			evaluation;

	public:
		Connection_Table();
	};

	struct  Cource_Selection
	{
		enum Action_Type { Hold , Find_Pass_Receiver , Pass ,
				   Push , Shoot };

		Action_Type	type;
		int		player_number;
	};

private:
	mutable	std::vector< Connection_Table >	table;
	mutable	Cource_Selection		result;
	mutable	int				first_best_player;
	mutable	bool	table_prepared;
	mutable	bool	search_finished;

private:
		void	prepare_table( const Field_Recog_Interface & ) const;

		void	prepare_result( const Field_Recog_Interface &  f ,
					int  recursive_level ,
					int  from ) const;

		double	recursive_search( const Field_Recog_Interface &  f ,
					  int  recursive_level ,
					  int  from ,
					  Cource_Selection *  result ) const;

		double	static_evaluation( const Field_Recog_Interface &  f ,
					   int  player_number ) const;

protected:
	virtual	double	pass_check( const Field_Recog_Interface &  f ,
				    const Player_Reference &  from ,
				    const Player_Reference &  to ) const;

	virtual	bool	shoot_check( const Field_Recog_Interface &  f ,
				     const Player_Reference &  from ) const;

	virtual	bool	push_check( const Field_Recog_Interface &  f ,
				    const Player_Reference &  from ) const;

	virtual	double	free_radius_check( const Field_Recog_Interface &  f ,
					   const Player_Reference &  pl )
									const;
public:
		 Pass_Cource_Table();
	virtual	~Pass_Cource_Table();

	virtual	const std::vector< Connection_Table > &
			get_table( const Field_Recog_Interface & ) const;

	virtual	Cource_Selection	get_first_selection
					  ( const Field_Recog_Interface & )
									const;

	virtual	Cource_Selection	get_result
					  ( const Field_Recog_Interface & )
									const;

	virtual	ref_count_ptr<Soccer_Action>
			get_action( const Field_Recog_Interface & ) const;

};


#endif	/* PASS_COURCE_TABLE_H_INCLUDED */
